import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;

/**
 * The {@link CrossNamespaceCopier} copies object trees like the
 * {@link EcoreUtil.Copier} but automatically crosses name spaces while copying.
 * For every object to be copied the name space of the class of the source
 * object will be mapped to a target name space and an object instantiated from
 * a class with the same name as the source class.
 * 
 * The {@link CrossNamespaceCopier} is intended for static use.
 * 
 * @author Enrico Schnepel
 */
@SuppressWarnings("serial")
public final class CrossNamespaceCopier extends NullAwareCopier {
	/**
	 * A thread-safe instance mapping
	 * 
	 * @see #getInstance()
	 */
	private final static Map<Thread, CrossNamespaceCopier> INSTANCE = new HashMap<Thread, CrossNamespaceCopier>();

	private final Map<EClass, EClass> classCache = new HashMap<EClass, EClass>();
	private final Map<String, String> namespaceMappings = new HashMap<String, String>();

	/**
	 * The class {@link CrossNamespaceCopier} has a static interface. Please use
	 * the {@link #crossNsCopy()} methods.
	 */
	private CrossNamespaceCopier() {
	}

	// @name Life cycle functions
	// @{

	/**
	 * Initializes a thread specific static instance for use with
	 * {@link #crossNsCopy(EObject)}.
	 * 
	 * @see #crossNsReleaseStaticInstance()
	 */
	public static void crossNsInitStaticInstance() {
		INSTANCE.put(Thread.currentThread(), new CrossNamespaceCopier());
	}

	/**
	 * Initializes a name space mapping in the current instance. Needed for
	 * {@link #crossNsCopy(EObject)}
	 * 
	 * @param sourceNamespace
	 *            the source name space
	 * @param targetNamespace
	 *            the target name space
	 */
	public static void crossNsSetupNamespaceMapping(String sourceNamespace,
			String targetNamespace) {
		CrossNamespaceCopier inst = getInstance();
		if (null == inst) {
			crossNsInitStaticInstance();
			inst = getInstance();
		}
		inst.namespaceMappings.put(sourceNamespace, targetNamespace);
	}

	/**
	 * Resolves the static instance according to the current thread.
	 * 
	 * @return the {@link CrossNamespaceCopier} instance
	 */
	private static CrossNamespaceCopier getInstance() {
		return INSTANCE.get(Thread.currentThread());
	}

	/**
	 * Releases the static instance associated with the current thread.
	 */
	public static void crossNsReleaseStaticInstance() {
		INSTANCE.remove(Thread.currentThread());
	}

	// @}

	// @name Operations
	// @{

	/**
	 * Helper function to copy just one object without having a static instance
	 * initialized. The source name space is automatically determined from the
	 * object's class.
	 * 
	 * @param o
	 *            the Object to copy
	 * @param targetNamespace
	 *            the target name space
	 * @return the copied object
	 * 
	 * @see #crossNsCopy(EObject)
	 */
	public static EObject crossNsCopy(EObject o, String targetNamespace) {
		String sourceNamespace = o.eClass().getEPackage().getNsURI();

		return crossNsCopy(o, sourceNamespace, targetNamespace);
	}

	/**
	 * Helper function to copy just one object without having a static instance
	 * initialized.
	 * 
	 * @param o
	 *            the Object to copy
	 * @param sourceNamespace
	 *            the source name space
	 * @param targetNamespace
	 *            the target name space
	 * @return the copied object
	 * 
	 * @see #crossNsCopy(EObject)
	 */
	public static EObject crossNsCopy(EObject o, String sourceNamespace,
			String targetNamespace) {
		CrossNamespaceCopier inst = new CrossNamespaceCopier();

		EObject o2 = inst.copy(o);
		inst.copyReferences();

		return o2;
	}

	/**
	 * Copies the object while migrating involved name spaces. A static instance
	 * must have been initialized with {@link #crossNsInitStaticInstance()} and
	 * {@link #crossNsSetupNamespaceMapping(String, String)}.
	 * 
	 * @param o
	 * @return
	 */
	public static EObject crossNsCopy(EObject o) {
		CrossNamespaceCopier inst = getInstance();

		EObject o2 = inst.copy(o);
		inst.copyReferences();

		return o2;
	}

	// @}

	// @name Algorithm functions
	// @{

	/**
	 * Implements the algorithm for migrating the objects in an object tree
	 * between name spaces.
	 */
	@Override
	protected EClass getTarget(EClass sourceClass) {
		EClass targetClass = classCache.get(sourceClass);
		if (null == targetClass) {
			String sourcePackageNS = sourceClass.getEPackage().getNsURI();
			String targetPackageNS = sourcePackageNS;
			for (Map.Entry<String, String> nsMapping : namespaceMappings
					.entrySet()) {
				if (sourcePackageNS.startsWith(nsMapping.getKey())) {
					targetPackageNS = sourcePackageNS.replace(nsMapping
							.getKey(), nsMapping.getValue());
				}
			}
			if (sourcePackageNS != targetPackageNS) {
				EPackage targetPackage = EPackage.Registry.INSTANCE
						.getEPackage(targetPackageNS);
				if (null != targetPackage) {
					targetClass = (EClass) targetPackage
							.getEClassifier(sourceClass.getName());
				}
			} else {
				targetClass = sourceClass;
			}
			classCache.put(sourceClass, targetClass);
		}
		return targetClass;
	}

	/**
	 * Implements the algorithm for migrating structural features in an object
	 * tree between name spaces.
	 */
	@Override
	protected EStructuralFeature getTarget(EStructuralFeature sourceFeature) {
		EClass sourceClass = sourceFeature.getEContainingClass();
		EClass targetClass = getTarget(sourceClass);
		EStructuralFeature targetFeature = targetClass
				.getEStructuralFeature(sourceFeature.getName());
		if ((null != targetFeature)
				&& (targetFeature.getEType().getName() != sourceFeature
						.getEType().getName())) {
			targetFeature = null;
		}
		return targetFeature;
	}

	// @}
}
